"""
**Case/Control Matching functions**

Contains all functions that implement the :ref:`maximizeControls` data preparation tool.
"""

import pandas as pd
from hopcroftkarp import HopcroftKarp
import numpy as np

CATEGORICAL_DATA = '675161f1c87ff2648c61ff1c57c780f2'


def generate_row_query(keys, deltas, tr):
	q = []
	for i, dt in enumerate(deltas):
		key = keys[i]
		is_categorical = dt == CATEGORICAL_DATA
		if is_categorical:
			part = '=='.join([key, tr[key].__repr__()])
		else:
			structure = ['abs(', key, '-', tr[key], ')', '<=', dt]
			part = ''.join([str(x) for x in structure])
		q.append(part)
	return '&'.join(q)


def get_options(targets, controls, keys, deltas):
	tt = targets[keys]
	c = controls[keys]
	matching = {}
	if len(c) > len(tt):
		for i in tt.index:
			tr = tt.loc[i]
			control_query = generate_row_query(keys, deltas, tr)
			matches = c.query(control_query).index
			# matching[i] = matches.drop_duplicates().tolist()
			matching[i] = set(matches)
	else:
		for i in c.index:
			tr = c.loc[i]
			target_query = generate_row_query(keys, deltas, tr)
			matches = tt.query(target_query).index
			# matching[i] = matches.drop_duplicates().tolist()
			matching[i] = set(matches)

	return matching


def output_matches(path, outputfile, data, all_used, success, goal, matched):
	new_data = data[data.index.isin(all_used)]
	print('---')

	if not success:
		print("Could not match data 1-%d, using the maximum number of matches found by the approximation algorithm" % goal)
		print("Matched data 1-%0.3f" % matched)
	else:
		print("Matched data 1-%s" % (matched))

	new_data.to_csv(path / outputfile, index=False)
	print("New group file in %s" % (path / outputfile))

	return


def control_match(path, input, output, keys, deltas, condition='genotype', goal=1):
	"""
	Estimate an optimal case-control mapping.

	Match cases/controls (defined by value of ``condition``\ ) with a ``goal``\ :1 ratio
	based on similarity of ``keys``\ . Specify tolerance interval for ``keys`` via ``deltas``\ .
	For an exact match, specify a delta of 0. The order of
	``deltas`` values must match the order of the ``keys``\ .

	Optimal matches are estimated via the HopcroftKarp algorithm. A new group file
	with all matched subjects is saved to ``output``\ . The explicit matching of
	cases to controls is saved to ``output``\ __matched_pairs.csv

	:param path: path to input and output
	:param input: name of file containing group data
	:param output: name of output matched group file
	:param keys: comma-separated list of matching criteria (columns in input file)
	:param deltas: comma-separated list of tolerance intervals for the matching criteria
	:param condition: Field denoting group assignments [default: genotype]
	:param goal: n, target matching ratio (control:case => n:1) [default: 1]

	:type path: pathlib Path
	:type input: str
	:type output: str
	:type keys: str
	:type deltas: str
	:type condition: str
	:type goal: int

	:returns: None

	.. note:: If the matching algorithm cannot achieve the specified matching ratio, it will issue a warning and report the achieved ratio.

	"""
	# Reformat arguments
	keys = keys.replace(" ", "").split(',')
	deltas = deltas.replace(" ", "").split(',')
	deltas = [CATEGORICAL_DATA if x == '' else float(x) for x in deltas]

	# save original goal value
	orig_goal = goal

	# Assign new value for outputfile
	if output is None:
		inname = input.strip('.csv')
		output = inname + '__matched.csv'
		match_file = path / (inname + '__matched_pairs.csv')
	else:
		outname = output.strip('.csv')
		match_file = path / (outname + '__matched_pairs.csv')

	# Read data from the provided input file
	data = pd.read_csv(path / input)

	# Assert that all of the provided matching keys are present in the data
	for key in keys:
		assert key in data.columns, '%s is not a column in the input file (%s)' % (key, input)
	# Assert that condition column is present in the data
	assert condition in data.columns, 'Specified condition (%s) is not a column in the input file (%s)' % (condition, input)
	# Assert that condition column contains only '1' and '0'
	condition_vals = np.unique(data[condition])
	assert len(condition_vals) == 2, 'There are %d values (should only be 2) in the specified condition column (%s) in the input file (%s)' % (len(condition_vals), condition, input)
	for val in [0, 1]:
		assert val in condition_vals, 'The value %d is missing from the condition column (%s) in the input file (%s)' % (val, condition, input)


	# Separate patients and controls
	match_by_group0 = len(data[data[condition] == 1]) > len(data[data[condition] == 0])
	if match_by_group0:
		print('There are more cases (%s=1) than controls (%s=0) -- matching by controls' %(condition, condition))
		targets = data[data[condition] == 0].copy()
		controls = data[data[condition] == 1].copy()
	else:
		print('There are more controls (%s=0) than cases (%s=1) -- matching by cases' %(condition, condition))
		targets = data[data[condition] == 1].copy()
		controls = data[data[condition] == 0].copy()
	# save original number of targets (used to check ifany are dropped)
	orig_num_target = targets.shape[0]

	# create dictionary to store matching pairs
	targets['matching_ix'] = [[] for _ in range(targets.shape[0])]
	pairing = targets[['matching_ix']].to_dict(orient='index')

	cid = set()
	tid = set()
	set_num = 0
	while goal > 0:
		set_num += 1
		print('Getting match set %d' % set_num)
		matching = get_options(targets, controls, keys, deltas) # get all possible matches for each target
		matched = HopcroftKarp(matching).maximum_matching() # find optimal pairings
		# store matches
		for i in targets.index:
			if i in matched:
				cid.add(matched[i])
				tid.add(i)
				pairing[i]['matching_ix'].append(matched[i])
		# remove matched IDs from control pool
		rem_ids = set(controls.index).difference(cid)
		controls = controls.loc[rem_ids,:]
		goal = goal - 1

	final_ratio = float(len(cid)) / float(len(tid))
	all_used = cid.union(tid)

	print('Formatting matches')
	# convert pairing to dataframe & get matching ids from ix
	pairing_df = pd.DataFrame.from_dict(pairing, orient='index')
	# copy matching ix and remove extra columns
	targets.loc[pairing_df.index, 'matching_ix'] = pairing_df['matching_ix']
	# get match IDs from index
	get_ids = lambda i: list(data.loc[i['matching_ix'], 'id'].values)
	targets['matching_ids'] = targets.apply(get_ids, axis=1)
	# separate list of subject IDs & get matching subject's info
	cols = keys[:]
	for i in range(0,orig_goal):
		match_col = 'match'+str(i+1)
		cols.append(match_col)
		expand = lambda x: pd.Series(x['matching_ids'])
		try:
			targets[match_col] = targets.apply(expand, axis=1)[i]
			for key in keys:
				match_info = pd.merge(targets[[match_col]], data[['id', key]], left_on=match_col, right_on='id')
				match_info.rename(columns={key: match_col + '_' + key}, inplace=True)
				targets = pd.merge(targets, match_info.drop(columns='id'), on=match_col,how='left')
				cols.append(match_col + '_' + key)
		except: # no matches found for set i
			targets[match_col] = np.nan

	cols.insert(0, 'id')
	cols.insert(1, 'genotype')
	# export matching pairs
	print('Saving case/control mapping to %s' %match_file)
	targets.to_csv(match_file,index=False, columns=cols)

	if len(tid) != orig_num_target:
		g = 'controls' if match_by_group0 else 'cases'
		print('WARNING: Some %s were dropped during matching (%d out of %d %s remain)' %(g,len(tid),orig_num_target,g))

	if final_ratio == orig_goal:
		matching_success = 1
	else:
		matching_success = 0

	output_matches(path, output, data, all_used, matching_success, orig_goal, final_ratio)

	return



from pyProWAS import *
import os
import numpy as np
import pandas as pd

reg_type = 0
str_reg_type = "log"
path = "/nfs/share5/clineci/DownSyndrome/experiments/prowas_test/"
filename = "cpts_age.csv"
groupfile = "group.csv"
phewas_cov = ''
outfile = 'feature_matrix.csv'
covariates = 'SEX'
str_thresh_type = "fdr"
thresh_type = 1

"""
# gen_ftype = reg_type
phenotypes = get_input(path, filename,reg_type)
genotypes = get_group_file(path, groupfile)
fm = generate_feature_matrix(genotypes, phenotypes, reg_type)

print("Saving feature matrices to %s" % (path + outfile))

np.savetxt(path + 'agg_measures_' + outfile, fm[0],delimiter=',')
print("...")
np.savetxt(path + 'icd_age_' + outfile, fm[1],delimiter=',')
print("...")
np.savetxt(path + 'phewas_cov_' + outfile, fm[2],delimiter=',')

regressions = run_phewas(fm, genotypes, covariates,reg_type)

print("Saving regression data to %s" % (path + 'regressions.csv'))
header = ','.join(['str_reg_type', str_reg_type, 'group', groupfile]) + '\n'
f = open(os.sep.join([path, 'regressions.csv']), 'w')
f.write(header)
regressions.to_csv(f,index=False)
f.close()

"""

regressions = pd.read_csv(path + 'regressions.csv',dtype={'PheWAS Code':str},skiprows=1)

print("creating plots")

# Check if an imbalance will be used

imbalances = get_imbalances(regressions)

y = regressions['"-log(p)"']
pvalues = regressions['p-val'].values

# Get the threshold type
if thresh_type == 0:
    thresh = get_bon_thresh(pvalues, 0.05)
elif thresh_type == 1:
    thresh = get_fdr_thresh(pvalues, 0.05)

thresh = 0.5
print('%s threshold: %0.5f'%(str_thresh_type,thresh))

try:
    regressions[['lowlim', 'uplim']] = regressions['Conf-interval beta'].str.split(',', expand=True)
    regressions['uplim'] = regressions.uplim.str.replace(']', '')
    regressions['lowlim'] = regressions.lowlim.str.replace('[', '')
    regressions = regressions.astype(dtype={'uplim':float,'lowlim':float})
    yb = regressions[['beta', 'lowlim', 'uplim']].values
    yb = yb.astype(float)
except Exception as e:
    print('Error reading regression file:')
    print(e)
    sys.exit()

save = path + 'plot.png'
file_name, file_format = os.path.splitext(save)
saveb = file_name + '_beta' + file_format
file_format = file_format[1:] # remove '.' from from first index
print("Saving plot to %s" % (save))

plot_manhattan(regressions, -math.log10(thresh), save=save, save_format=file_format)
plot_odds_ratio(regressions, -math.log10(thresh), save=saveb, save_format=file_format)


"""
**pyPheWAS Core version 2 (main pyPheWAS code)**

Contains all functions that drive the core PheWAS & ProWAS analysis tools.
"""

from collections import Counter
import getopt
import math
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
import os
import pandas as pd
import statsmodels.discrete.discrete_model as sm
import statsmodels.formula.api as smf
import matplotlib.lines as mlines
from tqdm import tqdm
import sys
import matplotlib


def print_start_msg():
	path = os.path.dirname(os.path.abspath(__file__))
	filename = os.sep.join([path, 'resources', 'pyPheWAS_start_msg.txt'])
	with open(filename, 'r') as msg_f:
		print('\n')
		for line in msg_f:
			print(line.strip('\n'))
	return

def get_codes(filename):
	"""
	Load PheWAS/ProWAS code map from the resources directory.

	:param filename: name of file in the resources folder to load
	:type filename: str

	:returns: code map from the given file
	:rtype: pandas DataFrame

	"""
	path = os.path.dirname(os.path.abspath(__file__))
	filename = os.sep.join([path, 'resources', filename])
	if 'icd9' in filename:
		data_col = 'ICD9'
		new_data_col = 'ICD_CODE'
		phecode_col = 'PheCode'
	elif 'icd10' in filename:
		data_col = 'ICD10'
		new_data_col = 'ICD_CODE'
		phecode_col = 'PheCode'
	else:
		data_col = 'cpt'
		new_data_col = 'CPT_CODE'
		phecode_col = 'prowas_code'
	try:
		phecode_map = pd.read_csv(filename, dtype={data_col:str, phecode_col:str})
		phecode_map.rename(columns={data_col:new_data_col},inplace=True)
		phecode_map.dropna(subset=[phecode_col], inplace=True)
	except Exception as e:
		print(e.args[0])
		print('Error loading phecode map : exiting pyPheWAS')
		sys.exit()

	return phecode_map


def get_group_file(path, filename):
	"""
	Read group data from the given file.
	Note: Any records with a null **id** are dropped.

	:param path: path to the file that contains the group data
	:param filename: name of the file that contains the group data.
	:type path: pathlib Path
	:type filename: string

	:returns: The data from the group file.
	:rtype: pandas DataFrame
	"""
	wholefname = path / filename
	genotypes = pd.read_csv(wholefname)
	genotypes = genotypes.dropna(subset=['id'])
	return genotypes


def get_icd_codes(path, filename, reg_type):
	"""
	Read ICD data from the given file and load it into a pandas DataFrame.

	ICD records are mapped to their correpsonding PheWAS Codes.
	The maximum age of each subject at each PheWAS Code is calculated and
	added to the DataFrame as the column *MaxAgeAtICD*. If ``reg_type`` = 2, the
	interval of time (years) over which a subject experiences each PheWAS Code
	is added as the column *duration*.

	:param path: path to the file that contains the phenotype data
	:param filename: name of the file that contains the phenotype data.
	:param reg_type: type of regression (0:binary, 1:count, 2:duration)
	:type path: pathlib Path
	:type filename: str
	:type reg_type: int

	:returns: data from the phenotype file.
	:rtype: pandas DataFrame

	"""

	wholefname = path / filename
	icdfile = pd.read_csv(wholefname,dtype={'ICD_CODE':str})
	icdfile['ICD_CODE'] = icdfile['ICD_CODE'].str.strip()
	icd_types = np.unique(icdfile['ICD_TYPE'])

	# check ICD types present in file
	if not all((icd_types == 9) | (icd_types == 10)):
		# TODO: add actual exception
		print('Found an ICD_TYPE that was not 9 or 10 - Please check phenotype file.\nExiting pyPheWAS')
		sys.exit()

	# merge with Phecode table, depends on which type(s) of ICD codes are in the icd file
	if icd_types.shape[0] == 2:
		print('Found both ICD-9 and ICD-10 codes.')
		icd9s = icdfile[icdfile['ICD_TYPE'] == 9]
		icd10s = icdfile[icdfile['ICD_TYPE'] == 10]
		phenotypes_9 = pd.merge(icd9s, icd9_codes,on='ICD_CODE')
		phenotypes_10 = pd.merge(icd10s, icd10_codes, on='ICD_CODE')
		phenotypes = phenotypes_9.append(phenotypes_10,sort=False)
	elif icd_types[0] == 9:
		print('Found only ICD-9 codes.')
		phenotypes = pd.merge(icdfile,icd9_codes,on='ICD_CODE')
	elif icd_types[0] == 10:
		print('Found only ICD-10 codes.')
		phenotypes = pd.merge(icdfile, icd10_codes, on='ICD_CODE')
	else:
		# TODO: add actual exception
		print('An issue occurred while parsing the ICD_TYPE column - Please check phenotype file.\nExiting pyPheWAS')
		sys.exit()

	if reg_type == 2:
		# pre-calculate durations for feature matrix step
		phenotypes['duration'] = phenotypes.groupby(['id', 'PheCode'])['AgeAtICD'].transform('max') - \
								 phenotypes.groupby(['id', 'PheCode'])['AgeAtICD'].transform('min') + 1

	# calculate max age at each ICD code
	phenotypes['MaxAgeAtICD'] = 0
	phenotypes['MaxAgeAtICD'] = phenotypes.groupby(['id', 'PheCode'])['AgeAtICD'].transform('max')

	return phenotypes


def get_cpt_codes(path, filename, reg_type):
	"""
	Read CPT data from the given file and load it into a pandas DataFrame.

	CPT records are mapped to their correpsonding ProWAS Codes.
	The maximum age of each subject at each ProWAS Code is calculated and
	added to the DataFrame as the column *MaxAgeAtCPT*. If ``reg_type`` = 2, the
	interval of time (years) over which a subject experiences each ProWAS Code
	is added as the column *duration*.

	:param path: path to the file that contains the phenotype data
	:param filename: name of the file that contains the phenotype data.
	:param reg_type: type of regression (0:binary, 1:count, 2:duration)
	:type path: pathlib Path
	:type filename: str
	:type reg_type: int

	:returns: data from the phenotype file.
	:rtype: pandas DataFrame

	"""

	wholefname = path / filename
	cptfile = pd.read_csv(wholefname,dtype={'CPT_CODE':str})
	cptfile['CPT_CODE'] = cptfile['CPT_CODE'].str.strip()
	phenotypes = pd.merge(cptfile, cpt_codes, on='CPT_CODE')

	if reg_type == 2:
		# pre-calculate durations for feature matrix step
		phenotypes['duration'] = phenotypes.groupby(['id', 'prowas_code'])['AgeAtCPT'].transform('max') - \
								 phenotypes.groupby(['id', 'prowas_code'])['AgeAtCPT'].transform('min') + 1

	# calculate max age at each ICD code
	phenotypes['MaxAgeAtCPT'] = 0
	phenotypes['MaxAgeAtCPT'] = phenotypes.groupby(['id', 'prowas_code'])['AgeAtCPT'].transform('max')

	return phenotypes



def generate_feature_matrix(genotypes, phenotype, reg_type, code_type, pheno_cov=None):
	"""
	Generates the feature matrix that will be used to run the PheWAS or ProWAS analysis. Feature matrix is 3xNxP,
	where N = number of subjects and P = number of PheWAS/ProWAS Codes.

	* Feature matrix [0] is the aggregate PheWAS/ProWAS matrix. It contains phenotype data aggregated across
	  each subject's record accoring to the specified ``reg_type``.
	* Feature matrix [1] is the age matrix, containing each subject's maximum age recorded for each phenotype.
	  For phenotypes absent in the subject's record, the subject's overall maximum age is recorded.
	* Feature matrix [2] is the phenotype covariate matrix. If ``pheno_cov`` is defined, it records whether or
	  not each subject has at least one instance of the specified phenotype in their record. Otherwise, it's a zero matrix.


	:param genotypes: group data
	:param phenotype: phenotype data retrieved from ``pyPheWAS.pyPhewasCorev2.get_icd_codes`` or ``pyPheWAS.pyPhewasCorev2.get_cpt_codes``
	:param reg_type: type of regression (0:binary, 1:count, 2:duration)
	:param code_type: type of EMR code ('ICD' or 'CPT')
	:param pheno_cov: *[optional]* PheWAS or ProWAS code to use as a covariate
	:type genotypes: pandas DataFrame
	:type phenotype: pandas DataFrame
	:type reg_type: int
	:type code_type: str
	:type pheno_cov: str

	:returns (feature_matrix, phenotype_header): (3xNxP feature matrix, PheWAS/ProWAS codes that correspond to the columns in the feature matrix)
	:rtype: (numpy array, list of str)

	.. note:: If ICD/CPT codes do not have a corresponding PheWAS/ProWAS code, they are removed by
           ``pyPheWAS.pyPhewasCorev2.get_icd_codes`` and ``pyPheWAS.pyPhewasCorev2.get_cpt_codes``. If a subject in
           the group *does not have any records in the phenotype DataFrame* (e.g. if none of their ICD/CPT codes had a
           correponding phenotype), they are *still included* in the feature matrix.

	"""

	# sort genotype and phenotype dataframes by 'id'
	print('Sorting phenotype data...')
	phenotype.sort_values(by='id', inplace=True)
	print('Sorting group data...')
	genotypes.sort_values(by='id', inplace=True)
	genotypes.reset_index(inplace=True,drop=True)

	# use phewas/prowas codes to make a dictionary of indices in the np array
	if code_type == 'CPT':
		phecode_col = 'prowas_code'
		empty_phewas_df = prowas_codes.set_index('prowas_code')
	else: # code_type == 'ICD'
		phecode_col = 'PheCode'
		empty_phewas_df = phewas_codes.set_index('PheCode')

	empty_phewas_df.sort_index(inplace=True)
	empty_phewas_df['np_index'] = range(0,empty_phewas_df.shape[0])
	np_index = empty_phewas_df['np_index'].to_dict()

	feature_matrix = np.zeros((3, genotypes.shape[0], empty_phewas_df.shape[0]), dtype=float)

	# make genotype a dictionary for faster access time
	genotypes_dict = genotypes.set_index('id').to_dict('index')

	exclude = []  # list of ids to exclude (in icd list but not in genotype list)
	last_id = ''  # track last id seen in icd list
	count = -1

	for _,event in tqdm(phenotype.iterrows(), desc="Processing "+code_type+"s", total=phenotype.shape[0]):
		curr_id = event['id']
		if not curr_id in genotypes_dict:
			if not curr_id in exclude:
				print('%s has records in phenotype file but is not in group file - excluding from study' % curr_id)
				exclude.append(curr_id)
			continue
		# check id to see if a new subject has been found
		if last_id != curr_id:
			count += 1
			while(curr_id != genotypes.loc[count,'id']): # subject at genotypes.loc[count] does not have ICD codes
				empty_id = genotypes.loc[count,'id']
				feature_matrix[1][count] = genotypes_dict[empty_id]['MaxAgeAtVisit']
				count += 1 # continue processing next subject
			last_id = curr_id  # reset last_id
			feature_matrix[1][count] = genotypes_dict[curr_id]['MaxAgeAtVisit']

		# get column index of event phecode
		phecode_ix = np_index[event[phecode_col]]
		# add the max at of that ICD code to the age feature matrix
		feature_matrix[1][count][phecode_ix] = event['MaxAgeAt'+code_type]
		# if using a phecode covariate and the event phecode is equal to the covariate phecode, set the fm2 to 1 for this subject
		if (pheno_cov is not None) & (event[phecode_col] == pheno_cov):
			feature_matrix[2][count][:] = 1

		if reg_type == 0:
			# binary aggregate: add a 1 to the phecode's column to denote that this subject had this phecode at some point
			feature_matrix[0][count][phecode_ix] = 1
		elif reg_type == 1:
			# count aggregate: add 1 to the phecode's column to find the total number of times this subject had this phecode
			feature_matrix[0][count][phecode_ix] += 1
		else:
			# dur aggregate: store the number of years between the first and last events with this phecode
			feature_matrix[0][count][phecode_ix] = event['duration'] # duration calculated in get_icd_codes()

	return feature_matrix,list(empty_phewas_df.index)



"""

Statistical Modeling

"""


def get_phenotype_info(p_index, code_type):
	"""
	Retrieve phenotype info.

	Return info for the phenotype code at the given index in the phenotype map. The phenotype map (PheWAS or ProWAS)
	depends on the value of ``code_type``. For ``code_type`` = 'ICD', PheWAS code info is returned. For ``code_type`` =
	'CPT', ProWAS code info is returned.

	The returned information consists of (in order) PheWAS/ProWAS code, Phenotype, and ICD-9/ICD-10 or CPT rollup (i.e. all
	ICD/CPT codes that map to the given PheWAS/ProWAS code)

	:param p_index: index of the desired phenotype code
	:param code_type: type of phenotype map to use
	:type p_index: int
	:type code_type: str

	:returns: list of phenotype info
	:rtype: list of str
	"""
	if code_type == 'ICD':
		p_code = phewas_codes.loc[p_index, 'PheCode']
		p_name = phewas_codes.loc[p_index, 'Phenotype']
		icd9_ix = icd9_codes['PheCode'] == p_code
		icd10_ix = icd10_codes['PheCode'] == p_code
	
		p_icd9_rollup = '/'.join(icd9_codes.loc[icd9_ix, 'ICD_CODE'].tolist())
		p_icd10_rollup = '/'.join(icd10_codes.loc[icd10_ix, 'ICD_CODE'].tolist())
		
		return [p_code, p_name, p_icd9_rollup, p_icd10_rollup]
	
	else:
		p_code = prowas_codes.loc[p_index, 'prowas_code']
		p_name = prowas_codes.loc[p_index, 'prowas_desc']
		cpt_ix = cpt_codes['prowas_code'] == p_code
		p_cpt_rollup = '/'.join(cpt_codes.loc[cpt_ix, 'CPT_CODE'].tolist())
		
		return [p_code, p_name, p_cpt_rollup]


def fit_pheno_model(genotypes, phen_vector1, phen_vector2, phen_vector3='', covariates='',
                         response='genotype',  lr=0, phenotype='', code_type='ICD'):
	"""
	Runs a logistic regression for a specific phenotype vector

	Compute a logistic regression of the form:

	:math:`Pr(response) \sim logit(phen\_vector1 + covariates)`

	``phen_vector1`` is a vector of phenotype aggregates of length N, where N = number of subjects.
	To use the age feature vector (``phen_vector2``), include 'MaxAgeAtICD' or 'MaxAgeAtCPT' in the ``covariates`` string.
	Other than 'MaxAgeAtICD' and 'MaxAgeAtCPT', all covariates and the response variable must be included in
	the group DataFrame.

	The returned results list consists of (in order) the -log\ :sub:`10`\ (p-value), p-value, beta, beta's confidence interval,
	and beta's standard error, estimated from the logit model for the ``phen_vector1`` variable.

	:param genotypes: group data
	:param phen_vector1: the aggregate phenotype vector
	:param phen_vector2: the maximum event age phenotype vector
	:param phen_vector3: *[optional]* the phenotype covariate vector
	:param covariates: *[optional]* covariates to include in the regressions separated by '+' (e.g. 'sex+ageAtDx')
	:param response: *[optional]* response variable in the logit model (default: *genotype*)
	:param lr: *[optional]* regularized maximum likelihood optimization flag (0 [default] = not regularized; 1 = regularized)
	:param phenotype: *[optional]* phenotype info [code, description] for this regression (used only for error handling)
	:param code_type: *[optional]*  EMR data type ('ICD' [default], or 'CPT')
	:type genotypes: pandas DataFrame
	:type phen_vector1: numpy array
	:type phen_vector2: numpy array
	:type phen_vector3: numpy array
	:type covariates: str
	:type response: str
	:type lr: int [0,1]
	:type phenotype: list of str
	:type code_type: str

	:returns: regression results
	:rtype: list

	"""

	data = genotypes.copy()
	data['y'] = phen_vector1 # aggregate phenotype data

	# append '+' to covariates (if there are any) -> makes definition of 'f' more elegant
	if covariates is not '':
		covariates = '+' + covariates

	# add MaxAgeAtEvent to data (if needed)
	if (code_type == 'ICD') and ('MaxAgeAtICD' in covariates):
		data['MaxAgeAtICD'] = phen_vector2
	elif (code_type == 'CPT') and ('MaxAgeAtCPT' in covariates):
		data['MaxAgeAtCPT'] = phen_vector2

	# add phewas_cov feature matrix to data & covariates (if needed)
	if phen_vector3.any():
		data['phe_cov'] = phen_vector3
		covariates = covariates + '+ phe_cov'

	# define model ('f') for the logisitc regression
	if lr == 1:
		predictors = covariates.replace(" ", "").split('+')
		predictors[0] = 'y'
		f = [response.strip(), predictors]
	else:
		f = response + '~ y' + covariates

	try:
		if lr == 0:
			# fit logit without regulatization
			model = smf.logit(f, data).fit(disp=False)
		elif lr == 1:
			# fit logit with regularization
			logit = sm.Logit(data[f[0]], data[f[1]])
			model = logit.fit_regularized(method='l1', alpha=0.1, disp=0, trim_mode='size', qc_verbose=0)
		else:
			linreg = smf.logit(f, data).fit(method='bfgs', disp=False)
		# get results
		p = model.pvalues.y
		beta = model.params.y
		conf = model.conf_int()
		conf_int = '[%s,%s]' % (conf[0]['y'], conf[1]['y'])
		stderr = model.bse.y
		reg_result = [-math.log10(p), p, beta, conf_int, stderr]  # collect results

	except ValueError as ve:
		print('\n')
		if phenotype is not '':
			print('ERROR computing regression for phenotype %s (%s)' %(phenotype[0],phenotype[1]))
		print(ve)
		print('lr = % d' %lr)
		reg_result = [np.nan, np.nan, np.nan, np.nan, np.nan]
	except Exception as e:
		print('\n')
		if phenotype is not '':
			print('ERROR computing regression for phenotype %s (%s)' %(phenotype[0],phenotype[1]))
		print(e)
		reg_result = [np.nan, np.nan, np.nan, np.nan, np.nan]
	return reg_result


def run_phewas(fm, genotypes, code_type, covariates='', response='genotype'):
	"""
	Run mass phenotype regressions

	Iterate over all PheWAS/ProWAS codes in the feature matrix, running a logistic regression of the form:

	:math:`Pr(response) \sim logit(phenotype\_aggregate + covariates)`

	``fm`` is a 3xNxP matrix, where N = number of subjects and P = number of PheWAS/ProWAS Codes; this should only
	be consutrcted by ``pyPheWAS.pyPhewasCorev2.generate_feature_matrix`` - otherwise results will be untrustworthy.
	To use the age feature matrix (``fm[1]``), include 'MaxAgeAtICD' or 'MaxAgeAtCPT' in the ``covariates`` string.
	Other than 'MaxAgeAtICD' and 'MaxAgeAtCPT', all covariates and the response variable must be included in
	the group DataFrame.

	The returned DataFrame includes the PheWAS/ProWAS code, Phenotype (code description, e.g. 'Pain in joint'),
	-log\ :sub:`10`\ (p-value), p-value, beta, beta's confidence interval, beta's standard error, and lists of the ICD-9/ICD-10 or
	CPT codes that map to the phenotype.

	:param fm: phenotype feature matrix derived via ``pyPheWAS.pyPhewasCorev2.generate_feature_matrix``
	:param genotypes: group data
	:param code_type:  type of EMR code ('ICD' or 'CPT')
	:param covariates: *[optional]* covariates to include in the regressions separated by '+' (e.g. 'sex+ageAtDx')
	:param response: *[optional]* response variable in the logisitc model (default: *genotype*)
	:type fm: numpy array
	:type genotypes: pandas DataFrame
	:type code_type: str
	:type covariates: str
	:type response: str

	:returns: regression results for each phenotype
	:rtype: pandas DataFrame


	.. note:: To prevent false positives & improve statistical power, regressions are only computed for
		  phenotypes which present for greater than 5 subjects in the cohort. Phenotypes which do not meet
		  this criteria are not included in the returned regression results.

	.. note:: For phenotypes that present in both the case (``response`` = 1) and control (``response`` = 0) groups, maximum
		  likelihood optimization is used to compute the logistic regression. For phenotypes that only present in
		  one of those groups, regularized maximum likelihood optimization is used.

	"""

	# sort group data by id
	print('Sorting group data...')
	genotypes.sort_values(by='id', inplace=True)
	genotypes.reset_index(inplace=True,drop=True)

	num_pheno = len(fm[0, 0])
	# check number of phenotypes in the feature matrix
	if code_type == 'ICD':
		assert num_pheno == phewas_codes.shape[0], "Expected %d columns in PheWAS feature matrix, but found %d. Please check the feature matrix" %(phewas_codes.shape[0], num_pheno)
	else:
		assert num_pheno == prowas_codes.shape[0], "Expected %d columns in ProWAS feature matrix, but found %d. Please check the feature matrix" %(prowas_codes.shape[0], num_pheno)

	# store all of the pertinent data from the regressions
	if code_type == 'ICD':
		out_cols = phewas_reg_cols
	else:
		out_cols = prowas_reg_cols
	regressions = pd.DataFrame(columns=out_cols)

	control = fm[0][genotypes[response] == 0, :]
	disease = fm[0][genotypes[response] == 1, :]
	# find all phecodes that only present for a single group (ie only controls or only case show the phecode) -> have to use regularization
	inds = np.where((control.any(axis=0) & ~disease.any(axis=0)) | (~control.any(axis=0) & disease.any(axis=0)))[0]
	for index in tqdm(range(num_pheno), desc='Running Regressions'):
		phen_info = get_phenotype_info(index, code_type)
		phen_vector1 = fm[0][:, index]
		phen_vector2 = fm[1][:, index]
		phen_vector3 = fm[2][:, index]
		# to prevent false positives, only run regressions if more than 5 records have positive values
		if np.where(phen_vector1 > 0)[0].shape[0] > 5:
			if index in inds: # decide whether or not to use regularization in logistic regression
				use_regular = 1
			else:
				use_regular = 0
			stat_info = fit_pheno_model(genotypes, phen_vector1, phen_vector2, phen_vector3, covariates,
			                                 response, phenotype = phen_info[0:2], lr=use_regular, code_type=code_type)
		else:
			# not enough samples to run regression
			stat_info = [np.nan, np.nan, np.nan, np.nan, np.nan]

		# save regression data
		info = phen_info[0:2] + stat_info + phen_info[2:]
			
		regressions.loc[index] = info

	return regressions.dropna(subset=['p-val']).sort_values(by='p-val') # sort by significance


"""

Result Visualization

"""


def get_bon_thresh(p_values, alpha=0.05):
	"""
	Calculate the Bonferroni multiple comparisons correction threshold for a list of p-values.

	Divide the power by the number of finite p-values values (all non-nan values).

	:param p_values: list of p-values
	:param alpha: the uncorrected significance level being used (default = 0.05)
	:type p_values: numpy array
	:type alpha: float

	:returns: The Bonferroni correction threshold
	:rtype: float

	"""
	return alpha / sum(np.isfinite(p_values))


def get_fdr_thresh(p_values, alpha=0.05):
	"""
	Calculate the false discovery rate (FDR) multiple comparisons correction threshold for a list of p-values.

	:param p_values: list of p-values
	:param alpha: the uncorrected significance level being used (default = 0.05)
	:type p_values: numpy array
	:type alpha: float

	:returns: The FDR correction threshold
	:rtype: float

	"""
	sn = np.sort(p_values)
	sn = sn[np.isfinite(sn)]
	for i in range(len(sn)):
		p_crit = alpha * float(i+1) / float(len(sn))
		if sn[i] <= p_crit:
			continue
		else:
			break
	return sn[i]



def get_bhy_thresh(p_values, alpha):
	# Deprecated
	"""
	Calculate the false discovery rate threshold.

	:param p_values: a list of p-values obtained by executing the regression
	:param alpha: the thershold power being used (usually 0.05)
	:type p_values: numpy array
	:type alpha: float

	:returns: the false discovery rate
	:rtype: float
	"""
	sn = np.sort(p_values)
	sn = sn[np.isfinite(sn)]
	# sn = sn[::-1]
	for i in range(len(sn)):
		p_crit = alpha * float(i+1) / (8.1 * float(len(sn)))
		if sn[i] <= p_crit:
			continue
		else:
			break
	return sn[i]


def get_imbalances(regressions):
	# Deprecated
	"""
	Generates a numpy array of the imbalances.

	For a value *x* where *x* is the beta of a regression:

	========= ====== =======================================================
	*x* < 0   **-1** The regression had a negative beta value
	*x* = nan **0**  The regression had a nan beta value (and a nan p-value)
	*x* > 0   **+1** The regression had a positive beta value
	========= ====== =======================================================

	These values are then used to get the correct colors using the imbalance_colors.

	:param regressions: DataFrame containing a variety of different output values from the regression performed. The only one used for this function are the 'beta' values.
	:type regressions: pandas DataFrame

	:returns: A list that is the length of the number of regressions performed. Each element in the list is either a -1, 0, or +1. These are used as explained above.
	:rtype: numpy array
	"""

	imbalance = np.array(regressions['beta'])
	imbalance[np.isnan(imbalance)] = 0
	imbalance[imbalance > 0] = 1
	imbalance[imbalance < 0] = -1
	return imbalance


def get_x_label_positions(categories, lines=True):
	# Deprecated
	"""
	This method is used get the position of the x-labels and the lines between the columns

	:param categories: list of the categories
	:param lines: a boolean which determines the locations returned (either the center of each category or the end)
	:type categories:
	:type lines: bool

	:returns: A list of positions
	:rtype: list of ints

	"""
	tt = Counter(categories)
	s = 0
	label_positions = []
	for _, v in tt.items():
		if lines:
			inc = v // 2
		else:
			inc = v
		label_positions.append(s + inc)
		s += v
	return label_positions


def plot_manhattan(regressions, thresh, code_type='ICD', show_imbalance=True, save='', save_format=''):
	"""
	Plots significant phenotype data on a Manhattan Plot.

	The significance of each phenotype (represented by :math:`-log_{10}(p)`\ ) is plotted along the
	y-axis, with phenotypes plotted along the x-axis.
	If ``save`` is provided, the plot is saved to a file; otherwise, the plot may be displayed with
	matplotlib.pyplot.show() after this function returns.


	:param regressions: dataframe containing the regression results
	:param thresh: p-value significance threshold
	:param code_type: type of EMR data ('ICD' or 'CPT')
	:param show_imbalance: boolean variable that determines whether or not to show imbalances on the plot (default True)
	:param save: the output file to save to (if empty, display the plot)
	:param save_format: format of the save file
	:type regressions: pandas DataFrame
	:type thresh: float
	:type code_type: str
	:type show_imbalance: boolean
	:type save: str
	:type save_format: str

	"""

	# Initialize figure
	fig = plt.figure(1)
	ax = plt.subplot(111)
	frame1 = plt.gca()

	# Merge regressions with phenotype data to get categories
	if code_type == 'ICD':
		pheno_codes = phewas_codes
		reg_key = 'PheWAS Code'
		map_key = 'PheCode'
		cat_key = 'category'
		pheno_name = 'Phenotype'
	else:
		pheno_codes = prowas_codes
		reg_key = 'ProWAS Code'
		map_key = 'prowas_code'
		cat_key = 'ccs'
		pheno_name = 'prowas_desc'
		
	regressions = pd.merge(regressions, pheno_codes, left_on=reg_key, right_on=map_key).sort_values(by=cat_key)

	# Plot all points w/ labels
	e = 1 # x-axis counter
	artists = []
	plt.ylabel('-log10(p)')

	ax.axhline(y=-math.log10(thresh), color='red', ls='dotted')  # plot threshold

	for ix,data in regressions.iterrows():
		logp_ix = data['"-log(p)"']
		if  data['p-val'] < thresh:
			# determine marker type based on whether/not showing imbalance
			if show_imbalance:
				mew = 1.5
				if data['beta'] > 0: m = '+'
				else: m = '_'
			else:
				mew = 0.0
				m = 'o'
			# Plot PheCode data point & format PheCode label
			if code_type == 'ICD': c = cat_colors[data['category_string']]
			else: c = 'xkcd:aqua' # constant color
			ax.plot(e, logp_ix, m, color=c, fillstyle='full', markeredgewidth=mew)
			artists.append(ax.text(e, logp_ix, data[pheno_name], rotation=89, va='bottom', fontsize=6))
			e += 15

	# Category Legend
	if code_type == 'ICD':
		line1 = []
		box = ax.get_position()
		ax.set_position([box.x0, box.y0 + box.height * 0.05, box.width, box.height * 0.95])
		for lab in cat_colors.keys():
			line1.append(mlines.Line2D(range(1), range(1), color="white", marker='o', markerfacecolor=cat_colors[lab], label=lab))
		artists.append(ax.legend(handles=line1, bbox_to_anchor=(0.5, 0), loc='upper center', fancybox=True, ncol=4, prop={'size': 6}))

	# Plot x axis
	ax.axhline(y=0, color='black')
	frame1.axes.get_xaxis().set_visible(False)

	# Save the plot
	if save:
		plt.savefig(save,
					format = save_format,
					bbox_extra_artists = artists,
					bbox_inches ='tight',
					dpi = 300
					)
		plt.close()

	return


def plot_odds_ratio(regressions, thresh, code_type='ICD', save='', save_format='', label_loc="plot"):
	"""
	Plots significant phenotype data on a Log Odds Plot.

	The log odds value & confidence interval is plotted along the x-axis, with Phenotypes sorted by category
	plotted along the y-axis.
	If ``save`` is provided, the plot is saved to a file; otherwise, the plot may be displayed with
	matplotlib.pyplot.show() after this function returns.

	:param regressions: dataframe containing the regression results
	:param thresh: p-value significance threshold
	:param code_type: type of EMR data ('ICD' or 'CPT')
	:param save: the output file to save to (if empty, display the plot)
	:param save_format: format of the save file
	:param label_loc: where to plot Phenotype labels ["plot" (defulat) or "axis"]
	:type regressions: pandas DataFrame
	:type thresh: float
	:type code_type: str
	:type save: str
	:type save_format: str
	:type label_loc: str

	"""

	# Initialize figure
	fig = plt.figure(2)
	ax = plt.subplot(111)
	frame1 = plt.gca()

	# Merge regressions with phenotype data to get categories
	if code_type == 'ICD':
		pheno_codes = phewas_codes
		reg_key = 'PheWAS Code'
		map_key = 'PheCode'
		cat_key = 'category'
		pheno_name = 'Phenotype'
	else:
		pheno_codes = prowas_codes
		reg_key = 'ProWAS Code'
		map_key = 'prowas_code'
		cat_key = 'ccs'
		pheno_name = 'prowas_desc'
	
	regressions = pd.merge(regressions, pheno_codes, left_on=reg_key, right_on=map_key).sort_values(by=cat_key)


	# Plot all points w/ labels
	e = 1 # vertical index
	text_size = 6
	artists = []
	if label_loc == "axis":
		phecode_labels = []
		phecode_locs = []
	plt.xlabel('Log odds ratio')
	for ix, data in regressions.iterrows():
		beta_ix = data['beta']
		if  data['p-val'] < thresh:
			# Add Phecode label
			if label_loc == "plot":
				if beta_ix > 0: # only difference is ha (horizontal alignment)
					artists.append(ax.text(beta_ix, e, data[pheno_name], rotation=0, ha='left', fontsize=text_size))
				else:
					artists.append(ax.text(beta_ix, e, data[pheno_name], rotation=0, ha='right', fontsize=text_size))
			else: # location = "axis"
				phecode_labels.append(data[pheno_name])
				phecode_locs.append(e)

			# Plot Phecode Data
			if code_type == 'ICD': c = cat_colors[data['category_string']]
			else: c = 'xkcd:aqua' # constant color
			ax.plot(beta_ix, e, 'o', color=c, fillstyle='full', markeredgewidth=0.0)
			ax.plot([data['lowlim'], data['uplim']], [e, e], color=c)
			e += 15

	# Plot y axis
	ax.axvline(x=0, color='black')

	if label_loc == "axis":
		plt.yticks(phecode_locs,phecode_labels, ha='right',fontsize=text_size)
	else:
		frame1.axes.get_yaxis().set_visible(False)

	# Legend
	if code_type == 'ICD':
		line1 = []
		box = ax.get_position()
		ax.set_position([box.x0, box.y0 + box.height * 0.05, box.width, box.height * 0.95])
		for lab in cat_colors.keys():
			line1.append(mlines.Line2D(range(1), range(1), color="white", marker='o', markerfacecolor=cat_colors[lab], label=lab))
		artists.append(ax.legend(handles=line1, bbox_to_anchor=(0.5, -0.125), loc='upper center', fancybox=True, ncol=4, prop={'size': text_size}))


	# Save the plot
	if save:
		plt.savefig(save,
					format = save_format,
					bbox_extra_artists = artists,
					bbox_inches ='tight',
					dpi = 300
					)
		plt.close()

	return


def plot_volcano(regressions, code_type='ICD', save='', save_format=''):
	"""
	Plots all phenotype data on a Volcano Plot.

	The significance of each phenotype (represented by :math:`-log_{10}(p)`\ ) is plotted along the
	y-axis, with log odds value (effect size) plotted along the x-axis. To improve plot legibility, only those
	phenotypes which surpass the FDR/Bonferroni significance thresholds have labels displayed.
	If ``save`` is provided, the plot is saved to a file; otherwise, the plot may be displayed with
	matplotlib.pyplot.show() after this function returns.

	:param regressions: dataframe containing the regression results
	:param code_type: type of EMR data ('ICD' or 'CPT')
	:param save: the output file to save to (if empty, display the plot)
	:param save_format: format of the save file
	:type regressions: pandas DataFrame
	:type code_type: str
	:type save: str
	:type save_format: str

	"""

	# get thresh values
	bon = get_bon_thresh(regressions["p-val"].values, 0.05)
	fdr = get_fdr_thresh(regressions["p-val"].values, 0.05)

	# Initialize figure
	fig = plt.figure(3)
	ax = plt.subplot(111)
	frame1 = plt.gca()

	# Plot all points w/ labels
	artists = []
	plt.ylabel('-log10(p)')
	plt.xlabel('Log Odds Ratio')
	
	if code_type == 'ICD':
		pheno_name = 'PheWAS Name'
	else:
		pheno_name = 'ProWAS Name'

	for ix,data in regressions.iterrows():
		logp_ix = data['"-log(p)"']
		beta = data['beta']
		# determine marker color & label based on thresholds
		if  data['p-val'] < bon:
			c = 'gold'
			phe = data[pheno_name]
		elif data['p-val'] < fdr:
			c = 'midnightblue'
			phe = data[pheno_name]
		else:
			c = 'slategray'
			phe = ''

		# Plot PheCode data point & format PheCode label
		ax.plot(beta, logp_ix, 'o', color=c, fillstyle='full', markeredgewidth=0)
		artists.append(ax.text(beta, logp_ix, phe, rotation=45, va='bottom', fontsize=4))

	# Legend
	line1 = []
	box = ax.get_position()
	ax.set_position([box.x0, box.y0 + box.height * 0.05, box.width, box.height * 0.95])

	line1.append(mlines.Line2D(range(1), range(1), color="white", marker='o', markerfacecolor="gold", label="BonFerroni"))
	line1.append(mlines.Line2D(range(1), range(1), color="white", marker='o', markerfacecolor="midnightblue", label="FDR"))
	line1.append(mlines.Line2D(range(1), range(1), color="white", marker='o', markerfacecolor="slategray", label="Insignificant"))

	artists.append(ax.legend(handles=line1, loc='best', fancybox=True, ncol=4, prop={'size': 6}))

	# Save the plot
	if save:
		plt.savefig(save,
					format = save_format,
					bbox_extra_artists = artists,
					bbox_inches ='tight',
					dpi = 300
					)
		plt.close()

	return



"""

Housekeeping & Parameters

"""


def process_args(kwargs, optargs, *args):
	# Deprecated
	clean = np.vectorize(lambda x: x[x.rfind('-') + 1:] + '=')
	searchfor = clean(list(optargs.keys()))
	opts, rem = getopt.getopt(args, '', searchfor)
	assert len(rem) == 0, 'Unknown arguments included %s' % (str(rem))
	for option in opts:
		k, v = option
		kwargs[optargs[k]] = v

	return kwargs


def display_kwargs(kwargs):
	for k, v in kwargs.items():
		num_dots = 80 - len(str(k)) - len(str(v))
		print(str(k) + '.'*num_dots + str(v))


phewas_reg_cols = ['PheWAS Code',
				  'PheWAS Name',
				  '\"-log(p)\"',
				  'p-val',
				  'beta',
				  'Conf-interval beta',
				  'std_error',
				  'ICD-9',
				  'ICD-10']

prowas_reg_cols = ['ProWAS Code',
				  'ProWAS Name',
				  '\"-log(p)\"',
				  'p-val',
				  'beta',
				  'Conf-interval beta',
				  'std_error',
				  'CPT']

cat_colors = {'other': 'gold',
               'circulatory system': 'xkcd:bright red',
               'congenital anomalies': 'mediumspringgreen',
               'dermatologic': 'xkcd:dark peach',
               'digestive': 'yellowgreen',
               'endocrine/metabolic': 'darkred',
               'genitourinary': 'seagreen',
               'hematopoietic': 'orange',
               'infectious diseases': 'blue',
               'injuries & poisonings': 'slategray',
               'mental disorders': 'xkcd:hot pink',
               'musculoskeletal': 'darkkhaki',
               'neoplasms': 'xkcd:bluish',
               'neurological': 'xkcd:purplish pink',
               'pregnancy complications': 'peachpuff',
               'respiratory': 'xkcd:carolina blue',
               'sense organs': 'darkviolet',
               'symptoms': 'aqua'}

old_plot_colors = {'-': 'gold',
			   'circulatory system': 'red',
			   'congenital anomalies': 'mediumspringgreen',
			   'dermatologic': 'maroon',
			   'digestive': 'green',
			   'endocrine/metabolic': 'darkred',
			   'genitourinary': 'black',
			   'hematopoietic': 'orange',
			   'infectious diseases': 'blue',
			   'injuries & poisonings': 'slategray',
			   'mental disorders': 'fuchsia',
			   'musculoskeletal': 'darkgreen',
			   'neoplasms': 'teal',
			   'neurological': 'midnightblue',
			   'pregnancy complications': 'gold',
			   'respiratory': 'brown',
			   'sense organs': 'darkviolet',
			   'symptoms': 'darkviolet'}

imbalance_colors = {
	0: 'white',
	1: 'deepskyblue',
	-1: 'red'
}
regression_map = {
	'log': 0,
	'lin': 1,
	'dur': 2
}
threshold_map = {
	'bon': 0,
	'fdr': 1,
	'custom':2
}

# make plot text editable
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype'] = 42


# load ICD maps (pyPheWAS)
icd9_codes = get_codes('phecode_map_v1_2_icd9.csv')
# icd9_codes = get_codes('phecode_map_v1_1_icd9.csv') # original ICD-PheCode mapping
icd10_codes = get_codes('phecode_map_v1_2_icd10_beta.csv')

phewas_codes = icd9_codes.append(icd10_codes[['PheCode','Phenotype','category','category_string']])
phewas_codes = phewas_codes[['PheCode','Phenotype','category','category_string']].dropna()
phewas_codes.drop_duplicates(subset='PheCode',inplace=True)
phewas_codes.sort_values(by=['PheCode'], inplace=True)
phewas_codes.reset_index(inplace=True, drop=True)

# load CPT maps (pyProWAS)
cpt_codes = get_codes('prowas_codes.csv')
prowas_codes = cpt_codes[['prowas_code','prowas_desc','ccs','CCS Label']].drop_duplicates(subset='prowas_code')
prowas_codes.sort_values(by=['prowas_code'], inplace=True)
prowas_codes.reset_index(inplace=True,drop=True)



"""
**EMR Data Censoring Function**

Contains source code for :ref:`censorData` tool.
"""

import pandas as pd
import numpy as np

def censor_diagnosis(genotype_file, phenotype_file, final_pfile, final_gfile, efield, delta_field=None, start_time=np.nan, end_time=np.nan):
	"""
	Specify a range of ages for censoring event data, such that ``efield`` ages are
	censored to the range

	        :math:`start \leq efield \leq end`

	Instead of censoring based on absolute age, you may also censor with respect to
	another data field using the ``delta_field``. If specified, the data is
	censored based on the *interval between* ``delta_field`` and ``efield``:

	        :math:`start \leq deltafield - efield \leq end`.

	Censored event data is saved to ``final_pfile``. Subjects with data remaining
	after censoring are saved to ``final_gfile``.

	:param genotype_file: path to input group file
	:param phenotype_file: path to input phenotype file
	:param final_pfile: path to output group file
	:param final_gfile: path to output group file
	:param efield: name of field in the phenotype file to be censored
	:param delta_field: name of field to censor with respect to (i.e. interval between ``delta_field`` and ``efield``) [default: None]
	:param start_time: start time for censoring in years [default: None]
	:param end_time: end time for censoring in years [default: None]

	:type genotype_file: str
	:type phenotype_file: str
	:type final_pfile: str
	:type final_gfile: str
	:type efield: str
	:type delta_field: str
	:type start_time: float
	:type end_time: float

	:returns: None

	.. note:: Either ``start_time`` and/or ``end_time`` must be given.

	"""
	# read files & check field names
	print('Reading input files')
	genotypes = pd.read_csv(genotype_file)
	phenotypes = pd.read_csv(phenotype_file)
	mg = pd.merge(phenotypes, genotypes, on='id')
	# assert specified fields exist
	assert efield in phenotypes, 'Specified efield (%s) does not exist in phenotype file' % efield

	# censor the data
	if delta_field is not None:
		# censor with respect to the interval between efield and delta_field
		assert delta_field in mg, 'Specified delta_field (%s) does not exist in phenotype or genotype file' % delta_field
		print('Censoring %s with respect to the interval between %s and %s' %(efield, efield, delta_field))
		mg['diff'] = mg[delta_field] - mg[efield]
		if np.isfinite(start_time) and np.isnan(end_time):
			print('Start time specified: keeping events with (%s - %s) >= %0.2f' % (delta_field,efield, start_time))
			# final = mg[(mg['diff']>=start_time)|(np.isnan(mg['diff']))] # old behavior keeps nans - for when controls don't have the delta_column?
			final = mg[(mg['diff'] >= start_time)]
		elif np.isnan(start_time) and np.isfinite(end_time):
			print('End time specified: keeping events with (%s - %s) <= %0.2f' % (delta_field, efield, end_time))
			# final = mg[(mg['diff']<=end_time)|(np.isnan(mg['diff']))] # old behavior keeps nans - for when controls don't have the delta_column?
			final = mg[(mg['diff'] <= end_time)]
		else:
			print('Start & End times specified: keeping events with %0.2 <= (%s - %s) <= %0.2f' % (start_time, delta_field, efield, end_time))
			# final = mg[(mg['diff']>=start_time)&(mg['diff']<=end_time)|(np.isnan(mg['diff']))]  # old behavior keeps nans - for when controls don't have the delta_column?
			final = mg[(mg['diff'] >= start_time) & (mg['diff'] <= end_time)]
	else:
		# censor efield ages based on how start/end times are specified
		if np.isfinite(start_time) and np.isnan(end_time):
			print('Start time specified: keeping events with %s >= %0.2f' %(efield,start_time))
			final = mg[mg[efield] >= start_time]
		elif np.isnan(start_time) and np.isfinite(end_time):
			print('End time specified: keeping events with %s <= %0.2f' %(efield,end_time))
			final = mg[mg[efield] <= end_time]
		else:
			print('Start & End times specified: keeping events with %0.2f <= %s <= %0.2f' %(start_time,efield,end_time))
			final = mg[(mg[efield] >= start_time) & (mg[efield] <= end_time)]

	# save results
	final_gp = final.drop_duplicates('id')
	print('%d (out of %d) subjects remain after censoring' %(final_gp.shape[0], genotypes.shape[0]))
	if 'genotype' in final_gp:
		num_case = final_gp[final_gp['genotype']==1].shape[0]
		num_ctrl = final_gp[final_gp['genotype']==0].shape[0]
		print('Cases: %d\nControls: %d' %(num_case,num_ctrl))

	print("Saving new genotype file to %s" % final_gfile)
	final_gp.to_csv(final_gfile, columns=genotypes.columns, index=False)
	print("Saving new phenotype file to %s" % final_pfile)
	final.to_csv(final_pfile, columns=phenotypes.columns, index=False)



from collections import Counter
import getopt
import math
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
import os
import pandas as pd
import scipy.stats
import statsmodels.api as sm
import statsmodels.formula.api as smf
from tqdm import tqdm
import matplotlib.lines as mlines


def get_codes():  # same
    """
    Gets the PheWAS codes from a local csv file and load it into a pandas DataFrame.

    :returns: All of the codes from the resource file.
    :rtype: pandas DataFrame

    """
    sep = os.sep
    path = os.path.dirname(os.path.abspath(__file__))
    filename = os.sep.join([path, 'resources', 'prowas_codes.csv'])
    return pd.read_csv(filename,dtype=str)


def get_group_file(path, filename):  # same
    """
    Read all of the genotype data from the given file and load it into a pandas DataFrame.

    :param path: The path to the file that contains the phenotype data
    :param filename: The name of the file that contains the phenotype data.
    :type path: string
    :type filename: string

    :returns: The data from the genotype file.
    :rtype: pandas DataFrame
    """
    wholefname = path + filename
    genotypes = pd.read_csv(wholefname)
    return genotypes


def get_input(path, filename, reg_type):  # diff -done - add duration
    """
    Read all of the phenotype data from the given file and load it into a pandas DataFrame.

    :param path: The path to the file that contains the phenotype data
    :param filename: The name of the file that contains the phenotype data.
    :type path: string
    :type filename: string

    :returns: The data from the phenotype file.
    :rtype: pandas DataFrame
    """
    wholefname = path + filename
    cptfile = pd.read_csv(wholefname)
    cptfile['cpt'] = cptfile['cpt'].str.strip()
    if reg_type == 0:
        phenotypes = pd.merge(cptfile, codes, on='cpt')
        phenotypes['MaxAgeAtCPT'] = 0
        phenotypes['MaxAgeAtCPT'] = phenotypes.groupby(['id', 'prowas_code'])['AgeAtCPT'].transform('max')
    else:
        """
        This needs to be changed, need to adjust for a variety of different naming conventions
        in the phenotype file, not simply 'AgeAtCPT', 'id', 'cpt', etc.
        Either we need to adjust for different names in the code, or state explicitly in the
        documentation that we cannot do things like this.
        """
        phenotypes = pd.merge(cptfile, codes, on='cpt')
        phenotypes['count'] = 0
        phenotypes['count'] = phenotypes.groupby(['id', 'prowas_code'])['count'].transform('count')
        phenotypes['duration'] = phenotypes.groupby(['id', 'prowas_code'])['AgeAtCPT'].transform('max') - \
                                 phenotypes.groupby(['id', 'prowas_code'])['AgeAtCPT'].transform('min') + 1
        phenotypes['MaxAgeAtCPT'] = 0
        phenotypes['MaxAgeAtCPT'] = phenotypes.groupby(['id', 'prowas_code'])['AgeAtCPT'].transform('max')
    return phenotypes


def generate_feature_matrix(genotypes, phenotypes, reg_type, phewas_cov=''):  # diff - done
    """
    Generates the feature matrix that will be used to run the regressions.

    :param genotypes:
    :param phenotypes:
    :type genotypes:
    :type phenotypes:

    :returns:
    :rtype:

    """
    pu=phenotypes[['id','prowas_code']].drop_duplicates()
    temp = pd.DataFrame(np.log2(pu['id'].drop_duplicates().count()/pu.groupby('prowas_code')['id'].count()).reset_index())
    temp.rename(columns={'id': 'idf'}, inplace=True)
    prowas_codes2 = pd.merge(prowas_codes, temp, on='prowas_code', how='left')

    feature_matrix = np.zeros((3, genotypes.shape[0], prowas_codes.shape[0]), dtype=int)
    count = 0;
    for i in tqdm(genotypes['id']):
        if reg_type == 0:
            temp = pd.DataFrame(phenotypes[phenotypes['id'] == i][['prowas_code', 'MaxAgeAtCPT']]).drop_duplicates()
            match = prowas_codes2['prowas_code'].isin(list(phenotypes[phenotypes['id'] == i]['prowas_code']))
            feature_matrix[0][count, match[match == True].index] = 1
            age = pd.merge(prowas_codes2, temp, on='prowas_code', how='left')['MaxAgeAtCPT']
            age[np.isnan(age)] = genotypes[genotypes['id'] == i].iloc[0]['MaxAgeAtVisit']
            assert np.all(np.isfinite(age)), "make sure MaxAgeAtVisit is filled"
            feature_matrix[1][count, :] = age
            if phewas_cov:
                feature_matrix[2][count, :] = int(phewas_cov in list(phenotypes[phenotypes['id'] == i]['prowas_code']))

        else:
            if reg_type == 1:
                temp = pd.DataFrame(
                    phenotypes[phenotypes['id'] == i][['prowas_code', 'MaxAgeAtCPT', 'count']]).drop_duplicates()
                cts = pd.merge(prowas_codes, temp, on='prowas_code', how='left')['count']
                cts[np.isnan(cts)] = 0
                feature_matrix[0][count, :] = cts
                age = pd.merge(prowas_codes2, temp, on='prowas_code', how='left')['MaxAgeAtCPT']
                age[np.isnan(age)] = genotypes[genotypes['id'] == i].iloc[0]['MaxAgeAtVisit']
                assert np.all(np.isfinite(age)), "make sure MaxAgeAtVisit is filled"
                feature_matrix[1][count, :] = age
                if phewas_cov:
                    feature_matrix[2][count, :] = int(
                        phewas_cov in list(phenotypes[phenotypes['id'] == i]['prowas_code']))

            elif reg_type == 2:
                temp = pd.DataFrame(
                    phenotypes[phenotypes['id'] == i][['prowas_code', 'MaxAgeAtCPT', 'count']]).drop_duplicates()
                temp = pd.merge(prowas_codes2, temp, on='prowas_code', how='left')
                tfidf=temp['count']*temp['idf']
                tfidf[np.isnan(tfidf)] = 0

                feature_matrix[0][count, :] = tfidf
                age = pd.merge(prowas_codes2, temp, on='prowas_code', how='left')['MaxAgeAtCPT']
                age[np.isnan(age)] = genotypes[genotypes['id'] == i].iloc[0]['MaxAgeAtVisit']
                assert np.all(np.isfinite(age)), "make sure MaxAgeAtVisit is filled"
                feature_matrix[1][count, :] = age
                if phewas_cov:
                    feature_matrix[2][count, :] = int(
                        phewas_cov in list(phenotypes[phenotypes['id'] == i]['prowas_code']))

        count += 1
    return feature_matrix


"""

Statistical Modeling

"""


def get_phewas_info(p_index):  # same
    """
    Returns all of the info of the phewas code at the given index.

    :param p_index: The index of the desired phewas code
    :type p_index: int

    :returns: A list including the code, the name, and the rollup of the phewas code. The rollup is a list of all of the cpt-9 codes that are grouped into this phewas code.
    :rtype: list of strings
    """
    p_code = prowas_codes.loc[p_index].prowas_code
    corresponding = codes[codes.prowas_code == p_code]

    p_name = corresponding.iloc[0].prowas_desc
    p_rollup = ','.join(codes[codes.prowas_code == p_code].cpt.tolist())
    return [p_code, p_name, p_rollup]


def calculate_odds_ratio(genotypes, phen_vector1, phen_vector2, covariates, lr=0, response='',
						 phen_vector3=''):  # diff - done
	"""
	Runs the regression for a specific phenotype vector relative to the genotype data and covariates.

	:param genotypes: a DataFrame containing the genotype information
	:param phen_vector1: a array containing the phenotype vector
	:param phen_vector2: a array containing the phenotype vector
	:param covariates: a string containing all desired covariates
	:type genotypes: pandas DataFrame
	:type phen_vector1: numpy array
	:type phen_vector2: numpy array
	:type covariates: string

	.. note::
		The covariates must be a string that is delimited by '+', not a list.
		If you are using a list of covariates and would like to convert it to the pyPhewas format, use the following::

			l = ['genotype', 'age'] # a list of your covariates
			covariates = '+'.join(l) # pyPhewas format

		The covariates that are listed here *must* be headers to your genotype CSV file.
	"""

	data = genotypes
	data['y'] = phen_vector1
	data['MaxAgeAtCPT'] = phen_vector2
	# f='y~'+covariates
	if covariates is not '':
		covariates = '+' + covariates
	if response:
		f = response + '~ y + genotype' + covariates
		if phen_vector3.any():
			data['phe'] = phen_vector3
			f = response + '~ y + phe + genotype' + covariates
	else:
		f = 'genotype ~ y' + covariates
		if phen_vector3.any():
			data['phe'] = phen_vector3
			f = 'genotype ~ y + phe' + covariates
	try:
		if lr == 0: # fit logit without regulatization
			logreg = smf.logit(f, data).fit(disp=False)
			p = logreg.pvalues.y
			odds = 0  #
			conf = logreg.conf_int()
			od = [-math.log10(p), p, logreg.params.y, '[%s,%s]' % (conf[0]['y'], conf[1]['y'])]
		elif lr == 1: # fit logit with regularization
			f1 = f.split(' ~ ')
			f1[1] = f1[1].replace(" ", "")
			logit = sm.Logit(data[f1[0].strip()], data[f1[1].split('+')])
			lf = logit.fit_regularized(method='l1', alpha=0.1, disp=0, trim_mode='size', qc_verbose=0)
			p = lf.pvalues.y
			odds = 0
			conf = lf.conf_int()
			od = [-math.log10(p), p, lf.params.y, '[%s,%s]' % (conf[0]['y'], conf[1]['y'])]
		else:
			linreg = smf.logit(f, data).fit(method='bfgs', disp=False)
			p = linreg.pvalues.y
			odds = 0
			conf = linreg.conf_int()
			od = [-math.log10(p), p, linreg.params.y, '[%s,%s]' % (conf[0]['y'], conf[1]['y'])]
	except ValueError as ve:
		print(ve)
		print('lr = % d' %lr)
		odds = 0
		p = np.nan
		od = [np.nan, p, np.nan, np.nan]
	except Exception as e:
		print(e)
		odds = 0
		p = np.nan
		od = [np.nan, p, np.nan, np.nan]
	return (odds, p, od)


def run_phewas(fm, genotypes, covariates, reg_type, response='', phewas_cov=''):  # same
	"""
	For each phewas code in the feature matrix, run the specified type of regression and save all of the resulting p-values.

	:param fm: The phewas feature matrix.
	:param genotypes: A pandas DataFrame of the genotype file.
	:param covariates: The covariates that the function is to be run on.
	:param reg_type: The covariates that the function is to be run on.
	:param response: The covariates that the function is to be run on.
	:param phewas_cov: The covariates that the function is to be run on.

	:returns: A tuple containing indices, p-values, and all the regression data.
	"""

	num_phecodes = len(fm[0, 0])
	thresh = math.ceil(genotypes.shape[0] * 0.03)
	# store all of the pertinent data from the regressions
	regressions = pd.DataFrame(columns=output_columns)
	control = fm[0][genotypes.genotype == 0, :]
	disease = fm[0][genotypes.genotype == 1, :]
	# find all phecodes that only present for a single genotype (ie only controls or only diseased show the phecode) -> have to use regularization
	inds = np.where((control.any(axis=0) & ~disease.any(axis=0)) | (~control.any(axis=0) & disease.any(axis=0)))[0]
	for index in tqdm(range(num_phecodes), desc='Running Regressions'):
		phen_vector1 = fm[0][:, index]
		phen_vector2 = fm[1][:, index]
		phen_vector3 = fm[2][:, index]
		# to prevent false positives, only run regressions if more than thresh records have positive values
		if np.where(phen_vector1 > 0)[0].shape[0] > thresh:
			if index in inds:
				res = calculate_odds_ratio(genotypes, phen_vector1, phen_vector2, covariates,
										   lr=1,
										   response=response,
										   phen_vector3=phen_vector3)
			else:
				res = calculate_odds_ratio(genotypes, phen_vector1, phen_vector2, covariates,
										   lr=0,
										   response=response,
										   phen_vector3=phen_vector3)

		else: # default (non-significant) values if not enough samples to run regression
			odds = 0
			p = np.nan
			od = [np.nan, p, np.nan, np.nan]
			res = (odds, p, od)

		# save all of the regression data
		phewas_info = get_phewas_info(index)
		stat_info = res[2]
		info = phewas_info[0:2] + stat_info + [phewas_info[2]]

		regressions.loc[index] = info

	return regressions.dropna(subset=['p-val']).sort_values(by='PheWAS Code')


def get_bon_thresh(normalized, power):  # same
    """
    Calculate the bonferroni correction threshold.

    Divide the power by the sum of all finite values (all non-nan values).

    :param normalized: an array of all normalized p-values. Normalized p-values are -log10(p) where p is the p-value.
    :param power: the threshold power being used (usually 0.05)
    :type normalized: numpy array
    :type power: float

    :returns: The bonferroni correction
    :rtype: float

    """
    return power / sum(np.isfinite(normalized))


def get_fdr_thresh(p_values, power):
    """
    Calculate the false discovery rate threshold.

    :param p_values: a list of p-values obtained by executing the regression
    :param power: the thershold power being used (usually 0.05)
    :type p_values: numpy array
    :type power: float

    :returns: the false discovery rate
    :rtype: float
    """
    sn = np.sort(p_values)
    sn = sn[np.isfinite(sn)]
    sn = sn[::-1]
    for i in range(len(sn)):
        thresh = 0.05 * i / len(sn)
        if sn[i] <= power:
            break
    return sn[i]


def get_imbalances(regressions):
    """
    Generates a numpy array of the imbalances.

    For a value *x* where *x* is the beta of a regression:

    ========= ====== =======================================================
    *x* < 0   **-1** The regression had a negative beta value
    *x* = nan **0**  The regression had a nan beta value (and a nan p-value)
    *x* > 0   **+1** The regression had a positive beta value
    ========= ====== =======================================================

    These values are then used to get the correct colors using the imbalance_colors.

    :param regressions: DataFrame containing a variety of different output values from the regression performed. The only one used for this function are the 'beta' values.
    :type regressions: pandas DataFrame

    :returns: A list that is the length of the number of regressions performed. Each element in the list is either a -1, 0, or +1. These are used as explained above.
    :rtype: numpy array
    """

    imbalance = np.array(regressions['beta'])
    imbalance[np.isnan(imbalance)] = 0
    imbalance[imbalance > 0] = 1
    imbalance[imbalance < 0] = -1
    return imbalance


def get_x_label_positions(categories, lines=True):  # same
    """
    This method is used get the position of the x-labels and the lines between the columns

    :param categories: list of the categories
    :param lines: a boolean which determines the locations returned (either the center of each category or the end)
    :type categories:
    :type lines: bool

    :returns: A list of positions
    :rtype: list of ints

    """
    tt = Counter(categories)
    s = 0
    label_positions = []
    for _, v in tt.items():
        if lines:
            inc = v // 2
        else:
            inc = v
        label_positions.append(s + inc)
        s += v
    return label_positions


def plot_manhattan(regressions, thresh, show_imbalance=True, save='', save_format=''):  # same
	"""
	Plots the data on a Manhattan Plot.

	:param regressions: dataframe containing the regression results
	:param thresh: the significance threshold
	:param save: the output file to save to (if empty, display the plot)
	:param show_imbalance: boolean variable that determines whether or not to show imbalances on the plot (default True)
	:type regressions: pandas DataFrame
	:type thresh: float
	:type save: str
	:type show_imbalance: boolean

	"""

	# Initialize figure
	fig = plt.figure(1)
	ax = plt.subplot(111)
	frame1 = plt.gca()

	# Merge regressions with Phewas data to get categories
	regressions = pd.merge(regressions, prowas_codes, left_on='PheWAS Code', right_on='prowas_code').sort_values(by='ccs')

	# Determine whether or not to show the imbalance.
	# show_imbalance = imbalances.size != 0

	# c = icd9_codes.loc[phewas_codes['index']]
	# c = c.reset_index()
	# idx = c.sort_values(by='category').index

	# Plot all points w/ labels
	e = 1
	artists = []
	plt.ylabel('-log10(p)')
	ax.axhline(y=thresh, color='red', ls='dotted')  # plot threshold
	for ix,data in regressions.iterrows():
		logp_ix = data['"-log(p)"']
		if  logp_ix > thresh:
			# determine marker type based on whether/not showing imbalance
			if show_imbalance:
				mew = 1.5
				if data['beta'] > 0: m = '+'
				else: m = '_'
			else:
				mew = 0.0
				m = 'o'
			# Plot PheCode data point & format PheCode label
			ax.plot(e, logp_ix, m, color="blue", fillstyle='full', markeredgewidth=mew)
			artists.append(ax.text(e, logp_ix, data['prowas_desc'], rotation=89, va='bottom', fontsize=6))
			e += 15

	# # Legend
	# line1 = []
	# box = ax.get_position()
	# ax.set_position([box.x0, box.y0 + box.height * 0.05, box.width, box.height * 0.95])
	# for lab in plot_colors.keys():
	# 	line1.append(mlines.Line2D(range(1), range(1), color="white", marker='o', markerfacecolor=plot_colors[lab], label=lab))
	# artists.append(ax.legend(handles=line1, bbox_to_anchor=(0.5, 0), loc='upper center', fancybox=True, ncol=4, prop={'size': 6}))

	# Plot x axis
	ax.axhline(y=0, color='black')
	frame1.axes.get_xaxis().set_visible(False)

	# If the imbalance is to be shown, draw lines to show the categories.
	# if show_imbalance:
	# 	for pos in linepos:
	# 		ax.axvline(x=pos, color='black', ls='dotted')

	# Save the plot
	if save:
		plt.savefig(save,format=save_format, bbox_extra_artists=artists, bbox_inches='tight')
		plt.clf()

	return


def plot_odds_ratio(regressions, thresh, show_imbalance=True, save='', save_format='', label_loc="plot"):  # same
	"""
	Plots the data on a Log Odds Plot.

	:param regressions: dataframe containing the regression results
	:param thresh: the significance threshold
	:param save: the output file to save to (if empty, display the plot)
	:param show_imbalance: boolean variable that determines whether or not to show imbalances on the plot (default True)
	:param label_loc: the output file to save to (if empty, display the plot)
	:type regressions: pandas DataFrame
	:type thresh: float
	:type save: str
	:type show_imbalance: boolean

	"""

	# Initialize figure
	fig = plt.figure(2)
	ax = plt.subplot(111)
	frame1 = plt.gca()

	# Merge regressions with Phewas data to get categories
	regressions = pd.merge(regressions, prowas_codes, left_on='PheWAS Code', right_on='prowas_code').sort_values(by='ccs')

	# determine whether or not to show imbalances
	# show_imbalance = imbalances.size != 0

	# Sort the phewas codes by category.
	# c = icd9_codes.loc[phewas_codes['index']]
	# c = c.reset_index()
	# idx = c.sort_values(by='category').index

	# Plot all points w/ labels
	e = 1  # vertical index
	ho = 0.025  # horizontal text offset
	vo = 1  # vertical text offset
	text_size = 6
	artists = []
	if label_loc == "axis":
		phecode_labels = []
		phecode_locs = []
	plt.xlabel('Log odds ratio')
	for ix, data in regressions.iterrows():
		beta_ix = data['beta']
		if data['"-log(p)"'] > thresh:
			# Add Phecode label
			if label_loc == "plot":
				if show_imbalance:
					if beta_ix > 0:
						artists.append(
							ax.text(beta_ix + ho, e + vo, data['prowas_desc'], rotation=0, ha='left', fontsize=text_size))
					else:
						artists.append(ax.text(beta_ix - ho, e + vo, data['prowas_desc'], rotation=0, ha='right',
						                       fontsize=text_size))
				else:
					artists.append(
						ax.text(beta_ix + ho, e + vo, data['prowas_desc'], rotation=0, va='bottom', fontsize=text_size))
			else:  # location = "axis"
				phecode_labels.append(data['prowas_desc'])
				phecode_locs.append(e)

			# Plot Phecode Data
			ax.plot(beta_ix, e, 'o', color="green", fillstyle='full', markeredgewidth=0.0)
			ax.plot([data['lowlim'], data['uplim']], [e, e], color="green")
			e += 15

	# Plot y axis
	ax.axvline(x=0, color='black')

	if label_loc == "axis":
		plt.yticks(phecode_locs, phecode_labels, ha='right', fontsize=text_size)
	else:
		frame1.axes.get_yaxis().set_visible(False)

	# Legend
	# line1 = []
	# box = ax.get_position()
	# ax.set_position([box.x0, box.y0 + box.height * 0.05, box.width, box.height * 0.95])
	# for lab in plot_colors.keys():
	# 	line1.append(
	# 		mlines.Line2D(range(1), range(1), color="white", marker='o', markerfacecolor=plot_colors[lab], label=lab))
	# artists.append(ax.legend(handles=line1, bbox_to_anchor=(0.5, -0.125), loc='upper center', fancybox=True, ncol=4,
	#                          prop={'size': text_size}))

	# If the imbalance is to be shown, draw lines to show the categories.
	# if show_imbalance:
	# 	for pos in linepos:
	# 		ax.axvline(x=pos, color='black', ls='dotted')

	# Save the plot
	if save:
		plt.savefig(save, format=save_format, bbox_extra_artists=artists, bbox_inches='tight')
		plt.clf()

	return

def process_args(kwargs, optargs, *args):
    clean = np.vectorize(lambda x: x[x.rfind('-') + 1:] + '=')
    searchfor = clean(list(optargs.keys()))
    opts, rem = getopt.getopt(args, '', searchfor)
    assert len(rem) == 0, 'Unknown arguments included %s' % (str(rem))
    for option in opts:
        k, v = option
        kwargs[optargs[k]] = v

    return kwargs


def display_kwargs(kwargs):
    print ("Arguments: ")
    for k, v in kwargs.items():
        left = str(k).ljust(30, '.')
        right = str(v).rjust(50, '.')
        print(left + right)


output_columns = ['PheWAS Code',
                  'PheWAS Name',
                  'p-val',
                  '\"-log(p)\"',
                  'beta',
                  'Conf-interval beta',
                  'cpt']

plot_colors = {'-': 'gold',
               'circulatory system': 'red',
               'congenital anomalies': 'mediumspringgreen',
               'dermatologic': 'maroon',
               'digestive': 'green',
               'endocrine/metabolic': 'darkred',
               'genitourinary': 'black',
               'hematopoietic': 'orange',
               'infectious diseases': 'blue',
               'injuries & poisonings': 'slategray',
               'mental disorders': 'fuchsia',
               'musculoskeletal': 'darkgreen',
               'neoplasms': 'teal',
               'neurological': 'midnightblue',
               'pregnancy complications': 'gold',
               'respiratory': 'brown',
               'sense organs': 'darkviolet',
               'symptoms': 'darkviolet'}
imbalance_colors = {
    0: 'white',
    1: 'deepskyblue',
    -1: 'red'
}
regression_map = {
    'log': 0,
    'lin': 1,
    'lind': 2
}
threshold_map = {
    'bon': 0,
    'fdr': 1
}
global codes, prowas_codes
codes = get_codes()
prowas_codes = codes[['prowas_code','prowas_desc','ccs','CCS Label']].drop_duplicates(subset='prowas_code')
prowas_codes.reset_index(inplace=True,drop=True)





from distutils.core import setup

setup(
  name = 'pyPheWAS',
  packages = ['pyPheWAS'], # this must be the same as the name above
  version = '3.1.0',
  description = 'MASI Lab Port of PheWAS into Python',
  author = 'MASI Lab',
  author_email = 'bennett.landman@vanderbilt.edu',
  url = 'https://github.com/MASILab/pyPheWAS', # use the URL to the github repo
  download_url = 'https://github.com/MASILab/pyPheWAS/tarball/0.1', # I'll explain this in a second
  keywords = ['PheWAS', 'ICD-9', 'ICD-10', 'EMR', 'CPT'], # arbitrary keywords
  classifiers = [],
  install_requires=[ 'numpy>=1.16.4',
                     'matplotlib',
                     'scipy>=1.2.1',
                     'pandas>=0.24.2',
                     'statsmodels>=0.10.1',
                     'hopcroftkarp',
                     'tqdm',
                     'pathlib',
                     ],
  package_data={
    '':['resources/*.csv']
  },
  scripts=['bin/pyPhewasLookup',
           'bin/pyPhewasModel',
           'bin/pyPhewasPlot',
           'bin/pyPhewasPipeline',
           'bin/pyProwasLookup',
           'bin/pyProwasModel',
           'bin/pyProwasPlot',
           'bin/pyProwasPipeline',
           'bin/censorData',
           'bin/convertEventToAge',
           'bin/createGenotypeFile',
           'bin/maximizeControls',
           'bin/mergeGroups'
           ],
)
